use crate::{
    expr::{Expr, ToExpr},
    intern::Interned,
    source_location::SourceLocation,
    ty::{
        impl_match_values_as_self, CanonicalType, CanonicalTypeKind, CanonicalValue, Connect, Type,
        TypeEnum, Value, ValueEnum,
    },
};
use bitvec::prelude::BitSlice;
use std::{
    fmt,
    sync::atomic::{AtomicU64, Ordering},
};

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
#[non_exhaustive]
pub struct HitUndeducedType;

impl HitUndeducedType {
    /// if either input is [`Ok(false)`][Ok], return [`Ok(false)`][Ok],
    /// otherwise if either input is [`Err(_)`][Err], return [`Err(_)`][Err],
    /// otherwise return [`Ok(true)`][Ok].
    pub fn reduce_and(l: Result<bool, Self>, r: Result<bool, Self>) -> Result<bool, Self> {
        match (l, r) {
            (Ok(false), _) | (_, Ok(false)) => Ok(false),
            (Err(e), _) | (_, Err(e)) => Err(e),
            (Ok(true), Ok(true)) => Ok(true),
        }
    }
}

impl fmt::Display for HitUndeducedType {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("encountered a not-yet-deduced type")
    }
}

impl std::error::Error for HitUndeducedType {}

#[derive(Copy, Debug, Clone, PartialEq, Eq, Hash)]
pub struct UndeducedType {
    id: u64,
    mask_ty_id: u64,
    source_location: SourceLocation,
}

impl UndeducedType {
    #[track_caller]
    pub fn new() -> Self {
        Self::new_with_loc(SourceLocation::caller())
    }
    pub fn new_with_loc(source_location: SourceLocation) -> Self {
        static NEXT_ID: AtomicU64 = AtomicU64::new(0);
        let id = NEXT_ID.fetch_add(2, Ordering::Relaxed);
        let mask_ty_id = id + 1;
        Self {
            id,
            mask_ty_id,
            source_location,
        }
    }
    pub fn id(self) -> u64 {
        self.id
    }
}

impl<T> Connect<T> for UndeducedType {}

impl Type for UndeducedType {
    type CanonicalType = UndeducedType;
    type Value = Deduce;
    type CanonicalValue = Deduce;
    type MaskType = UndeducedType;
    type MaskValue = Deduce;
    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        Self {
            id: self.mask_ty_id,
            mask_ty_id: self.mask_ty_id,
            source_location: self.source_location,
        }
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        self.source_location
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::Deduce(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }
}

impl CanonicalType for UndeducedType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::Deduce;
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub enum Deduce {}

impl ToExpr for Deduce {
    type Type = UndeducedType;

    fn ty(&self) -> Self::Type {
        match *self {}
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        match *self {}
    }
}

impl Value for Deduce {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        *self
    }
}

impl CanonicalValue for Deduce {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        match *this {}
    }

    fn to_bits_impl(this: &Self) -> Result<Interned<BitSlice>, HitUndeducedType> {
        match *this {}
    }
}
